"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.observe = exports.TEARDOWN_DELAY = void 0;
const retil_support_1 = require("retil-support");
const source_1 = require("./source");
exports.TEARDOWN_DELAY = 10;
function observe(observable, parentAct) {
    const asyncActs = new Set();
    const callbacks = [];
    let actDeferred = null;
    let actDepth = 0;
    let error = null;
    let nextVector = null;
    let sealed = false;
    // A null indicates that we don't know the current value. An empty array
    // indicates that we should have a subscription.
    let vector = null;
    let subscription = null;
    const observableSubscribe = typeof observable === 'function' ? observable : observable.subscribe;
    const get = () => {
        subscribeIfRequired();
        if (error) {
            throw error.value;
        }
        if (!vector) {
            vector = [];
        }
        // This must be called *after* the snapshot is created, as it checks if
        // the snapshot is a deferred.
        scheduleTeardownIfRequired();
        return vector;
    };
    const subscribe = (change, seal) => {
        if (sealed) {
            if (seal) {
                seal();
            }
            return retil_support_1.noop;
        }
        if (error) {
            return retil_support_1.noop;
        }
        const pair = [change, seal];
        callbacks.push(pair);
        subscribeIfRequired();
        return () => {
            const index = callbacks.indexOf(pair);
            if (index !== -1) {
                callbacks.splice(index, 1);
                scheduleTeardownIfRequired();
            }
        };
    };
    const handleValue = (value) => {
        const receivedVector = Array.isArray(value) ? value : [value];
        const latestVector = nextVector || vector;
        if (sealed ||
            !subscription ||
            (latestVector && (0, retil_support_1.areArraysShallowEqual)(receivedVector, latestVector))) {
            return;
        }
        nextVector = receivedVector;
        if (!actDepth) {
            commit();
        }
    };
    const handleSeal = () => {
        if (!subscription) {
            return;
        }
        if (nextVector) {
            commit();
        }
        if (vector === null || !vector.length) {
            handleError(new Error('Attempted to seal an observe() with no value'));
        }
        sealed = true;
        // Tear down the subscription *without* removing the value
        const unsubscribe = subscription.unsubscribe;
        subscription = null;
        try {
            unsubscribe();
        }
        catch (_a) { }
        callbacks.slice().forEach(([, seal]) => {
            if (seal) {
                seal();
            }
        });
        callbacks.length = 0;
    };
    const handleError = (err) => {
        if (sealed || !subscription) {
            return;
        }
        error = { value: err };
        notifySubscribers();
        teardownSubscription();
        callbacks.length = 0;
        if (actDeferred) {
            actDeferred.reject(err);
        }
    };
    const notifySubscribers = () => {
        // Some observables will immediately synchronously call `next` during
        // subscribe to let us know the current value. In this case, we'll
        // skip notifying subscribers, as they can get the value if they need
        // it.
        if (subscription && !subscription.isSubscribing) {
            callbacks.slice().forEach(callChangeListener);
        }
    };
    const subscribeIfRequired = () => {
        if (sealed) {
            return;
        }
        if (subscription) {
            subscription.count++;
            if (subscription.teardownTimeout) {
                clearTimeout(subscription.teardownTimeout);
            }
        }
        else if (!error) {
            subscription = {
                count: 1,
                isSubscribing: true,
            };
            const unsubscribeFunctionOrObject = observableSubscribe(handleValue, handleError, handleSeal);
            const unsubscribe = typeof unsubscribeFunctionOrObject === 'function'
                ? unsubscribeFunctionOrObject
                : unsubscribeFunctionOrObject.unsubscribe;
            if (subscription) {
                subscription.isSubscribing = false;
                subscription.unsubscribe = unsubscribe;
            }
            else {
                unsubscribe();
            }
        }
    };
    const scheduleTeardownIfRequired = () => {
        // Don't teardown if we've emitted an empty vector, without having emited
        // a follow-on non-empty vector.
        if (subscription &&
            --subscription.count === 0 &&
            (!vector || vector.length > 0)) {
            scheduleTeardown(subscription);
        }
    };
    const scheduleTeardown = (subscription) => {
        // TODO: use requestIdleCallback instead if possible.
        subscription.teardownTimeout = setTimeout(teardownSubscription, exports.TEARDOWN_DELAY);
    };
    const teardownSubscription = () => {
        // Avoid teardown if we've since resubscribed
        if (subscription && subscription.count === 0) {
            const unsubscribe = subscription.unsubscribe;
            nextVector = null;
            vector = null;
            subscription = null;
            try {
                unsubscribe();
            }
            catch (_a) { }
        }
    };
    const commit = () => {
        if (sealed || !subscription)
            return;
        const actDeferredCopy = actDeferred;
        actDeferred = null;
        if (nextVector && (!vector || !(0, retil_support_1.areArraysShallowEqual)(vector, nextVector))) {
            vector = nextVector;
            nextVector = null;
            notifySubscribers();
        }
        if (actDeferredCopy) {
            actDeferredCopy.resolve();
        }
        // If we skipped a teardown due to an unresolved promise, we can now finish
        // it off.
        if (subscription &&
            subscription.count === 0 &&
            vector &&
            vector.length > 0) {
            scheduleTeardown(subscription);
        }
    };
    // TODO: queue until the first subscriber is added - include listeners to the
    // promise returned by this function as listeners. All callbacks become async
    // if there's no listeners, and require any value to be cleared.
    const act = (callback) => {
        if (error) {
            throw error.value;
        }
        if (sealed) {
            return callback();
        }
        const isTopLevelAct = ++actDepth === 1;
        const batch = (actDeferred = actDeferred || new retil_support_1.Deferred());
        subscribeIfRequired();
        const result = parentAct ? parentAct(callback) : callback();
        if ((0, retil_support_1.isPromiseLike)(result)) {
            const asyncAct = Promise.resolve(result).then(() => {
                --actDepth;
                scheduleTeardownIfRequired();
                asyncActs.delete(asyncAct);
                if (asyncActs.size === 0) {
                    commit();
                }
            }, handleError);
            asyncActs.add(asyncAct);
            // Temporarily clear the result while waiting for the async action.
            if (vector && vector.length > 0) {
                // Save the current snapshot in case nothing happens
                nextVector = nextVector || vector;
                vector = [];
                notifySubscribers();
            }
        }
        else {
            --actDepth;
            scheduleTeardownIfRequired();
            if (isTopLevelAct && asyncActs.size === 0) {
                commit();
                return result;
            }
        }
        return batch.promise.then(() => result);
    };
    return [[get, subscribe], source_1.identitySelector, act];
}
exports.observe = observe;
const callChangeListener = ([listener]) => {
    try {
        if (listener) {
            listener();
        }
    }
    catch (errorOrPromise) {
        // Given callbacks will call `getSnapshot()`, which often throws a promise,
        // let's ignore thrown promises so that the callback don't have to.
        if (!(0, retil_support_1.isPromiseLike)(errorOrPromise)) {
            throw errorOrPromise;
        }
    }
};
