/**
 * Note, there's no need for a version, as snapshots are immutable. A change
 * in snapshot indicates a new version.
 *
 * The only exception is on suspense/error. In these cases, `useSource` can
 * always use the same constant value as the version, as React doesn't care
 * what the suspense is or what the error is -- only that an suspense/error
 * was thrown.
 */
export declare type Source<TSelection, TValue = any> = readonly [
    SourceCore<TValue>,
    SourceSelect<TSelection, TValue>,
    SourceAct
];
export declare type SourceCore<TValue = any> = readonly [
    SourceGetVector<TValue>,
    SourceSubscribe
];
export declare type GettableSource<TSelection, TValue = any> = readonly [
    SourceCore<TValue>,
    SourceSelect<TSelection, TValue>,
    SourceAct?
];
/**
 * Subscribe to notification of new snapshots being available.
 *
 * The first callback will be called at some point in the same tick after the
 * value has changed. Note that it will *not* necessarily be called once for
 * every change in value; it may skip values, but will always be called for the
 * last change in the tick.
 *
 * The second callback will be called when it's guaranteed that no future
 * changes will be made. This allows any subscribers to clean up their listener
 * functions to prevent memory leaks.
 */
export declare type SourceSubscribe = (change?: () => void, seal?: () => void) => SourceUnsubscribe;
export declare type SourceUnsubscribe = () => void;
export declare type SourceGetVector<TValue = any> = () => TValue[];
/**
 * Takes the value returned from a SourceGetVector function (i.e. the version
 * vector), and returns the part of the value that you're actually interested in
 * (i.e. the snapshot, typically the first item).
 *
 * Note that you can only select on a source with a non-empty version vector. If
 * the source currently has no version (i.e. it's in an error or busy state),
 * then there'll be no snapshot either.
 */
export declare type SourceSelect<TSelection, TValue = any> = (value: TValue) => TSelection;
/**
 * An optional act function, which if exists, will batch synchronous updates,
 * and suspend the source until asynchronous updates are complete.
 */
export declare type SourceAct = <TActResult>(callback: () => PromiseLike<TActResult> | TActResult) => PromiseLike<TActResult> | TActResult;
export declare function act<TSelection, TActResult>(source: Source<TSelection>, callback: () => PromiseLike<TActResult> | TActResult): Promise<TActResult>;
export declare function getSnapshot<TSelection>(source: GettableSource<TSelection>): TSelection;
export declare function getSnapshotPromise<TSelection, TItem = TSelection>([[getVector, subscribe], select,]: GettableSource<TSelection, TItem>): Promise<TSelection>;
export declare function getVector<T>([[getVector]]: readonly [
    SourceCore<T>,
    SourceSelect<any, any>?,
    SourceAct?
]): T[];
export declare function hasSnapshot([[getVector]]: readonly [
    SourceCore<unknown>,
    SourceSelect<any, any>?,
    SourceAct?
]): boolean;
export declare function subscribe([[, subscribe]]: Source<unknown>, onChange?: () => void, onSeal?: () => void): SourceUnsubscribe;
export declare const identitySelector: <U>(value: unknown) => U;
export declare const constant: <T>(value: T) => Source<T, T>;
export declare const constantVector: <T>(value: T[]) => Source<T, T>;
export declare const nullSource: Source<null, null>;
export declare const pendingSource: Source<any>;
